/****************************************************************

kernel/cpuid.c
CPU identification code for AWOS

See LICENSE file included with this source for copyright
information.

****************************************************************/

#include <kernel.h>
#include <cpuinfo.h>
#include <string.h>

char *KiCpuBrandStringIntel[] =
{ "none. (Unknown processor)",
"Intel(R) Celeron(R) processor (01).",
"Intel(R) Pentium(R) III processor (02).",
"Intel(R) Pentium(R) III Xeon(TM) processor (03).",
"Intel(R) Pentium(R) III processor (04).",
"reserved (05).",
"Mobile Intel(R) Pentium(R) III-M processor (06).",
"Mobile Intel(R) Celeron(R) processor (07).",
"Intel(R) Pentium(R) 4 processor (08).",
"Intel(R) Pentium(R) 4 processor (09).",
"Intel(R) Celeron(R) processor (0A).",
"Intel(R) Xeon(TM) processor (0B).",
"Intel(R) Xeon(TM) MP processor (0C).",
"reserved (0D).",
"Mobile Intel(R) Pentium(R) 4-M processor (0E).",
"Mobile Intel(R) Celeron(R) processor (0F).",
"reserved (10).",
"Mobile Genuine Intel(R) processor (11).",
"Intel(R) Celeron(R) M processor (12).",
"Mobile Intel(R) Celeron(R) processor (13).",
"Intel(R) Celeron(R) processor (14).",
"Mobile Genuine Intel(R) processor (15).",
"Intel(R) Pentium(R) M processor (16).",
"Mobile Intel(R) Celeron(R) processor (17)." };

void KiCpuIdIntel(struct KiCpuInfo *);
void KiCpuIdAmd(struct KiCpuInfo *);
int KiGetCpuClass();

void KiCpuId(struct KiCpuInfo *cpu)
{
	long eax, ebx, ecx, edx;
	
	printf("Running on ");
	
	switch(KiGetCpuClass())
	{
		case 1:
			strcpy("GenuineIntel", cpu->mfr);
			cpu->features.family = 1;
			panic(" an Intel 8086/8088 or compatible. Get a real computer.");
			/* NOTREACHED */
		case 2:
			strcpy("GenuineIntel", cpu->mfr);
			cpu->features.family = 2;
			panic(" an Intel 286 or compatible. We don't support such ancient garbage. Sorry.");
			/* NOTREACHED */
		case 3:
			strcpy("GenuineIntel", cpu->mfr);
			cpu->features.family = 3;
			printf(" an Intel386 or compatible.");
			return;
		case 4:
			strcpy("GenuineIntel", cpu->mfr);
			cpu->features.family = 4;
			printf(" an Intel486 below stepping 3.");
			return;
		case 5:
			/* fall through -- CPUID instruction supported */
			eax = ebx = ecx = edx = 0;
	};

	// XXX - Should go somewhere else.	
	asm("CPUID" :
		"=a"(eax), "=b"(ebx), "=c"(ecx), "=d"(edx) :
		"a"(0)
	);
	
	memcpy(cpu->mfr, &ebx, sizeof(ebx));
	memcpy(cpu->mfr + sizeof(ebx), &edx, sizeof(edx));
	memcpy(cpu->mfr + sizeof(ebx) + sizeof(edx), &ecx, sizeof(ecx));
	cpu->mfr[12] = 0;
	
	if(!strcmp(cpu->mfr, "GenuineIntel"))
	{
		printf("an Intel");
		KiCpuIdIntel(cpu);
	} else if(!strcmp(cpu->mfr, "AuthenticAMD")) {
		printf("an AMD");
		KiCpuIdAmd(cpu);
	} else if(!strcmp(cpu->mfr, "CyrixInstead")) {
		printf("a Cyrix.\n");
	} else if(!strcmp(cpu->mfr, "CentaurHauls")) {
		printf("a Centaur.\n");
	} else if(!strcmp(cpu->mfr, "SiS SiS SiS ")) {
		printf("an SiS.\n");
	} else if(!strcmp(cpu->mfr, "NexGenDriven")) {
		printf("a NexGen.\n");
	} else if((!strcmp(cpu->mfr, "GenuineTMx86")) || (!strcmp(cpu->mfr, "TransmetaCPU"))) {
		printf("a Transmeta.\n");
	} else if(!strcmp(cpu->mfr, "RiseRiseRise")) {
		printf("a Rise.\n");
	} else if(!strcmp(cpu->mfr, "UMC UMC UMC ")) {
		printf("a UMC.\n");
	} else if(!strcmp(cpu->mfr, "Geode by NSC")) {
		printf("a National Semiconductor Geode.\n");
	} else {
		printf("an unknown manufacturer: '%s'.\n", cpu->mfr);
	};
	
	return;
};

char *CPUUnknownMessage = "Unknown (report f/m/s and model name to team)";

#define I486MODELS 6
char *Intel486Models[] = 
{
	"DX", "DX", "SX", "DX2 / DX2 Overdrive / 487",
	"SL", "SX2", "DX2 (Write-Back enhanced", "DX4 / DX4 Overdrive"
};

#define I586MODELS 4
char *Intel586Models[] = 
{
	"Never used", "(60/66)", "(75/133)", "Overdrive for Intel486", "with MMX"
};

#define I686MODELS 15
char *Intel686Models[] = 
{
	"Never used", "Pentium Pro", "Never used", "Pentium II Klamath",
	"Never used", "Pentium II Deschutes", "Celeron Mendocino", "Pentium III Katmai",
	"Pentium III Coppermine", "Pentium M Banias", "Pentium III Xeon",  "Pentium III Tualatin",
	"Unknown", "Pentium/Celeron M model D", "Core Solo/Duo", "probably Core 2"
};

void KiCpuIdIntel(struct KiCpuInfo *cpu)
{
	long eax, ebx, ecx, edx;
	
	asm("CPUID" :
		"=a"(eax), "=b"(ebx), "=c"(ecx), "=d"(edx) :
		"a"(1)
	);
	
	//0:3 - Stepping
	//4:7 - Model
	//8:11 - Family
	//12:13 - Processor Type
	//16:19 - Extended Model
	//20:27 - Extended Family
	cpu->features.stepping	= (eax   )  & 15;
	cpu->features.model	= (eax>>4)  & 15;
	cpu->features.family	= (eax>>8)  & 15;
	cpu->features.type	= (eax>>12) &  3;
	cpu->features.ext_model	= (eax>>16) & 15;
	cpu->features.ext_family= (eax>>20) & 63;
	cpu->features.brand	= (ebx    ) & 0xFFFF;
	cpu->features.c_flags	= ecx;
	cpu->features.d_flags	= edx;

	switch(cpu->features.family)
	{
		case 4:
			//486
			printf("486 %s (%i/%i/%i).\n", (cpu->features.model > I486MODELS ? CPUUnknownMessage : Intel486Models[cpu->features.model]), cpu->features.family, cpu->features.model, cpu->features.stepping);
			return;
		case 5:
			//586 / Pentium
			printf(" Pentium %s (%d/%d/%d).\n", (cpu->features.model > I586MODELS ? CPUUnknownMessage : Intel586Models[cpu->features.model]), cpu->features.family, cpu->features.model, cpu->features.stepping);
			break;
		case 6:
			//686 / Pentium Pro / Pentium II / Celeron / Pentium III
			// don't waste instructions with the embedded if, we have all 16 possibilities.
			printf(" %s (%d/%d/%d).\n", Intel686Models[cpu->features.model], cpu->features.family, cpu->features.model, cpu->features.stepping);
			break;
		case 15:
			//786 / NetBurst / Pentium 4 / Celeron
			switch(cpu->features.model)
			{
				case 0:
					printf(" Pentium 4/Xeon/Celeron (f%d, m%d, s%d).\n", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 1:
					printf(" Pentium 4/Xeon/Celeron (Willamette) (f%d, m%d, s%d).\n", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 2:
					printf(" Pentium 4/Xeon/Celeron (Northwood) (f%d, m%d, s%d).\n", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				default:
					printf(" NetBurst unknown (f%d, m%d, s%d).\n", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
			};
			return;
		default:
			printf(" unknown (f%d, m%d, s%d).\n", cpu->features.family, cpu->features.model, cpu->features.stepping);
			return;
	};
};

void KiCpuIdAmd(struct KiCpuInfo *cpu)
{
	long eax, ebx, ecx, edx;
	
	asm("CPUID" :
		"=a"(eax), "=b"(ebx), "=c"(ecx), "=d"(edx) :
		"a"(1)
	);
	
	cpu->features.stepping	= (eax   ) & 15;
	cpu->features.model	= (eax>>4) & 15;
	cpu->features.family	= (eax>>8) & 15;
	
	switch(cpu->features.family)
	{
		case 4:
			dprintf(" Am486 or Am5x86 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
			break;
		case 5:
			//Pentium-compatibles
			switch(cpu->features.model)
			{
				case 0:
					dprintf(" K5 model 0 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 1:
					dprintf(" K5 model 1 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 2:
					dprintf(" K5 model 2 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 3:
					dprintf(" K5 model 3 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 6:
					dprintf(" K6 model 6 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 7:
					dprintf(" K6 model 7 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 8:
					dprintf(" K6-2 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 9:
					dprintf(" K6-III (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				default:
					dprintf(" K5/K6 unknown (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
			};
			break;
		case 6:
			//Athlon
			switch(cpu->features.model)
			{
				case 1:
					dprintf(" K7 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 2:
					dprintf(" Athlon model 2 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 3:
					dprintf(" Duron model 3 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 4:
					dprintf(" Athlon model 4 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 6:
					dprintf(" Athlon/Duron MP/XP (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				case 7:
					dprintf(" Duron model 7 (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
				default:
					dprintf(" Athlon/Duron unknown (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
					break;
			};
			break;
		default:
			dprintf(" unknown (f%d, m%d, s%d).", cpu->features.family, cpu->features.model, cpu->features.stepping);
			break;
	};
};

extern struct KiCpuInfo *cpuInfo;

BOOL CpuCheckFlag(long flag)
{
	// XXX only supports d_flags atm
	return (cpuInfo->features.d_flags & flag);
};
