/****************************************************************
 *
 * cons/vgactl.c
 * Text-mode VGA control driver for AWOS
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
****************************************************************/

/*! \file vgactl.c
 * Text-mode VGA control
 */

#include <kernel.h>
#include <string.h>
#include <cons.h>
#include <portio.h>
#include <stdint.h>
#include <wchar.h>

TTY_MAKEDRV_TEXT(
	cons_vga_80x25,
	
	VgaClass_pop,	VgaClass_push,	VgaClass_tseek,	VgaClass_tincx,
	VgaClass_tincy,	VgaClass_tdecx,	VgaClass_tdecy,	VgaClass_vscroll,
	VgaClass_putch,	VgaClass_cls
);

struct video_moder *vga_modelist,*vga_modelist_end;
struct video_moder vmode_02 = {
	video_modeset_vga13, NULL
};
struct video_moder vmode_01 = {
	video_modeset_text80x25, &vmode_02
};


/*! \fn void VgaUpdateCursor()
 * \brief Updates the current tty's cursor position.
 * \note This code would also work on CGA, MCGA, and EGA.
 * \return Nothing.
 */
void VgaUpdateCursor()
{
	int16_t curpos;
	
	/* curpos holds the current position of the cursor
	 * only do this math once to save cycles */
	curpos = ((currtty->ty * 80) + currtty->tx);

	/* set the VGA registers */
	outb(VGA_REG_IDX, VGA_REG_CURSORHI);
	outb(VGA_REG_DATA, curpos >> 8);
	outb(VGA_REG_IDX, VGA_REG_CURSORLO);
	outb(VGA_REG_DATA, curpos);
};

/*! \fn void VgaClearScreen()
 * \brief Clears the current tty's screen.
 * \return Nothing.
 */
void VgaClearScreen()
{
	currtty->driver->cls(currtty);
};

void VgaCtlInit(void)
{
	vga_modelist = &vmode_01;
	vga_modelist_end = &vmode_02;
};

/*! \fn void VgaClearScreen()
 * \brief Clears the selected tty's screen.
 * \return Nothing.
 */
void VgaClass_cls(console_t *tty)
{
	tty->tx = 0;
	tty->ty = 0;
	wmemset((uint16_t *)tty->currmem, 0x0720, tty->memlen >> 1);
	if(currtty == tty)
		VgaUpdateCursor();
};

/*! \fn void VgaSetConsole(console_t *tty)
 * \brief Attempts to set video mode from a tty.
 * \param tty tty to activate.
 * \return Nothing.
 */
void VgaSetConsole(console_t *tty)
{
	cons_currclass = tty->driver;
	currtty->driver->push(currtty);
	cons_currclass = currtty->driver;
	tty->driver->pop(tty);
	cons_currclass = tty->driver;
	currtty = tty;
};

/*! \fn console_t *VgaSetMode(int width, int height, int bpp)
 * \brief Attempts to obtain a tty from the first driver in a list.
 * \param width width
 * \param height height
 * \param bpp bpp
 * \return A new console_t * if it worked, NULL if it failed.
 */
console_t *VgaSetMode(int width, int height, int bpp)
{
	struct video_moder *vga;
	console_t *tty;
	vga = vga_modelist;
	while(vga != NULL)
	{
		if((tty = vga->fn(width, height, bpp)))
		{
			return tty;
		}
		vga = vga->n;
	};
	return NULL;
};

/*! \fn void VgaClass_pop(console_t *tty)
 * \brief "Pops" a text-mode console (view).
 * \param tty The tty to pop.
 * \return nothing
 * This function takes a text-mode console (tty), retrives its memory from
 * vritual-tty memory, writes it to Video RAM, and then sets its active flag.
 * Use only after VgaClass_push'ing the previously active console.
 */
void VgaClass_pop(console_t *tty)
{
	if(!(tty->flags & CONS_ACTIVE))
	{
		memcpy((void *)tty->vrammem, (void *)tty->ttymem, tty->memlen);
		tty->currmem = tty->vrammem;
		tty->flags |= CONS_ACTIVE | CONS_TEXTMODE | CONS_PERSIST;
	};
};

/*! \fn void VgaClass_push(console_t *tty)
 * \brief "Pushes" a text-mode console (un-view).
 * \param tty The tty to push.
 * \return nothing
 * This function takes a text-mode console (tty), writes its memory to the
 * virtual-tty memory, and unsets its active flag.  Use when switching a vt
 * before using VgaClass_pop.
 */
void VgaClass_push(console_t *tty)
{
	if(tty->flags & CONS_ACTIVE)
	{
		memcpy((void *)tty->ttymem, (void *)tty->vrammem, tty->memlen);
		tty->currmem = tty->ttymem;
		tty->flags &= ~CONS_ACTIVE;
	};
};

/*! \fn void VgaClass_tseek(console_t *tty, int x, int y)
 * \brief Seeks the console memory to the specified coordinate.
 * \param tty The tty to seek.
 * \param x The x coordinate.
 * \param y The y coordinate.
 * \return nothing
 * This function seeks the memory of tty to the position specified by x and y.
 */
void VgaClass_tseek(console_t *tty, int x, int y)
{
	tty->tx = x;
	tty->ty = y;
	tty->mem = tty->currmem + (y * tty->tline) + x * tty->fw;
};

void VgaClass_tincx(console_t *tty)
{
	tty->tx++;
	tty->mem += tty->fw;
};

void VgaClass_tdecx(console_t *tty)
{
	tty->tx--;
	tty->mem -= tty->fw;
};

void VgaClass_tincy(console_t *tty)
{
	tty->ty++;
	tty->mem += tty->tline;
};

void VgaClass_tdecy(console_t *tty)
{
	tty->ty--;
	tty->mem -= tty->tline;
};

/*! \fn void VgaClass_vscroll(console_t *tty)
 * \brief Scrolls tty up a line.
 * \param tty The tty to scroll.
 * \return nothing
 * This function takes tty and moves the contents of its memory up exactly one
 * line.  This has the effect of "scrolling up".
 */
void VgaClass_vscroll(console_t *tty)
{
	char *where;
	where = (void *)(tty->currmem + tty->memlen - tty->tline);
	memcpy((int8_t *)tty->currmem, (int8_t *)(tty->currmem + tty->tline), tty->memlen - tty->tline);
	wmemset((uint16_t *)where, 0x0720, tty->tline>>1);
	tty->ty--;
};

void VgaClass_putch(console_t *tty, char c)
{
	*(tty->mem) = c;
	*(tty->mem + 1) = tty->color;
	if(tty == currtty)
	{
		currtty->tx++;
		VgaUpdateCursor();
		currtty->tx--;
	}
};

