/***************************************************************
 *
 * arch/i386/fault.c
 * x86 fault code for AWOS / V86 monitor code
 *
 * See LICENSE file include with this source for copyright
 * information.
 * 
 ***************************************************************/

#include <kernel.h>
#include <string.h>
#include <stdlib.h>
#include <portio.h>
#include <sched.h>
#include <cons.h>	/* currtty */
#include <v86.h>	/* V86... */
#include <cpuspec.h>	/* struct regs */

#define	PGF_PRESENT	0x01
#define	PGF_WRITE	0x02
#define	PGF_USER	0x04
#define	PGF_RSVD	0x08
#define	PGF_IF		0x10

struct frame
{
	struct frame *next;
	void *address;
};

char decide(int);

char *exception_messages[] =
{
	"Division by Zero", "Debug", "Non-Maskable Interrupt", "Breakpoint",
	"INTO Detected Overflow", "Out of Bounds", "Invalid Opcode", "No Coprocessor",
	"Double Fault", "Coprocessor Segment Overrun", "Bad TSS", "Segment Not Present",
	"Stack Fault", "General Protection Fault", "Page Fault", "Unknown Interrupt",
	"Coprocessor Fault", "Alignment Check (486+)", "Machine Check (Pentium/586+)",
	"SIMD floating-point exception", "DOS syscall", "Reserved", "Reserved", "Reserved",
	"Reserved", "Reserved", "Reserved", "Reserved",
	"Reserved", "Windows syscall", "Reserved", "Reserved",
	"Kernel Detected Fault", "Reserved", "Reserved", "Reserved",
	"Reserved", "Reserved", "Reserved", "Reserved",
	"Reserved", "Reserved", "Reserved", "Reserved",
	"Reserved", "Reserved", "Reserved", "Reserved",
	"Reserved", "Reserved", "Reserved", "Reserved",
	"Reserved", "Reserved", "Reserved", "Reserved",
	"Reserved", "Reserved", "Reserved", "Reserved",
	"Reserved", "Reserved", "Reserved",
	"Deprecated V86 entry method called by code we forgot to update"
	// ^ Just covering all bases. This should be removed eventually.
	// If we forgot something, this'll pop up.
};

extern struct tss ktss;
void tr_update(short);
uint32_t __KiGetCR0(void);
uint32_t __KiGetCR2(void);
uint32_t __KiGetCR3(void);
uint16_t __KiGetTR(void);
void __KiRestartByJump(void);

#define V86F_OPSIZE 0x00000001

void __KiV86Return_main(void);

void KsV86Init(void)
{
	// Set up interrupts in real mode IDT.
	*((uint32_t *)(0x40*4))=0xFEEBCD40; // yield/csleep interrupt
	printf("v86: ready.\n");
};

void KiV86Return(struct regs *r)
{
	r->eip = (void *)(__KiV86Return_main);
	r->cs = 0x0008;
	r->eflags = 0; // Dear V86 code: don't set stupid flags. Love, GM.
};

struct regs *KiFaultHandler(struct regs *r)
{
	unsigned char *a, *b;
	struct frame { struct frame *prev; void *ip; } *frame = (struct frame *)r->ebp;
	int i;
	uint32_t v86f;
	uint16_t reps;
	int dir;
	struct tss *tss;
	BOOL fail = FALSE;
	/*if(r->int_no == 0x3F) {
		dprintf("v86: KiV86Enter via INT 0x3F\n");
		KiV86Enter(0x0800, 0x0000, 0xFFFF);
		dprintf("v86: Returning to 32-bit task\n");
		return r;
	} else*/ if (r->int_no == 0x01) {
		// Debug exceptions suck.
		if((r->eflags & 0x20000))
		{
			dprintf("v86 debug exception at %04X:%04X\n", r->cs, ((uint32_t)r->eip) & 0xFFFF);
			r->ss &= 0xFFFF;
			r->useresp -= 2;
			r->useresp &= 0xFFFF;
			*(uint16_t *)(r->useresp + (r->ss<<4)) = r->eflags & 0xFFFF;
			r->useresp -= 2;
			r->useresp &= 0xFFFF;
			*(uint16_t *)(r->useresp + (r->ss<<4)) = r->cs;
			r->useresp -= 2;
			r->useresp &= 0xFFFF;
			*(uint16_t *)(r->useresp + (r->ss<<4)) = (uint16_t)(uint32_t)(r->eip);
			r->cs = *(uint16_t *)4;
			r->eip = (void *)(uint32_t)(*(uint16_t *)6);
			r->eflags &= ~0x100; // Remove trap flag.
		}
	} else {
		if((r->eflags & 0x20000) && r->int_no == 0x0D) /* We don't want to crash all the time on V86. */
		{
			a = (void *)((uint32_t)r->eip + (r->cs<<4));
			v86f = 0;
			reps = 1;
			if(r->eflags & 0x0400)
				dir = -1;
			else
				dir = 1;
			//dprintf("#",*a);
			while(!fail)
			{
				//dprintf("%02X",*a);
				switch(*a)
				{
					case 0x66: // operand size flag
						v86f |= V86F_OPSIZE;
						break;
					case 0x6C: // INSB
						do {
							*(uint8_t *)((r->edi & 0xFFFF) + (r->ves<<4)) = inb(r->edx & 0xFFFF);
							r->edi += dir;
							reps--;
						} while (reps != 0);
						r->eip++;
						return r;
					case 0x6D: // INSW/INSD
						do {
							if(v86f & V86F_OPSIZE)
							{
								*(uint32_t *)((r->edi & 0xFFFF) + (r->ves<<4)) = ind(r->edx & 0xFFFF);
								r->edi += dir<<2;
							}
							else
							{
								*(uint16_t *)((r->edi & 0xFFFF) + (r->ves<<4)) = inw(r->edx & 0xFFFF);
								r->edi += dir<<1;
							};
							reps--;
						} while (reps != 0);
						r->eip++;
						return r;
					case 0x6E: // OUTSB
						do {
							outb(r->edx & 0xFFFF, *(uint8_t *)((r->esi & 0xFFFF) + (r->vds<<4)));
							r->esi += dir;
							reps--;
						} while (reps != 0);
						r->eip++;
						return r;
					case 0x6F: // OUTSW/OUTSD
						do {
							if(v86f & V86F_OPSIZE)
							{
								outd(r->edx & 0xFFFF, *(uint16_t *)((r->esi & 0xFFFF) + (r->vds<<4)));
								r->esi += dir<<2;
							}
							else
							{
								outw(r->edx & 0xFFFF, *(uint32_t *)((r->esi & 0xFFFF) + (r->vds<<4)));
								r->esi += dir<<1;
							};
							reps--;
						} while (reps != 0);
						r->eip++;
						return r;
					case 0x9C: // PUSHF
						r->useresp -= 2;
						*((uint16_t *)r->useresp) = (uint32_t)r->eflags & 0xFFFF;
						r->eip++;
						return r;
					case 0x9D: // POPF
						r->eflags = (r->eflags & 0xFFFF0000) | *((uint16_t *)r->useresp);
						r->useresp += 2;
						r->eip++;
						return r;
					case 0xCD: // INT 0x__
						b = (void *)(a[1] << 2);
						if((*(uint32_t *)b | 0xFF) == 0xFEEBCDFF)
						{
							if(*(uint32_t *)b == 0xFEEBCD40) /* Yield/csleep */
							{
								r->eip += 2;
								if(r->eax & 0xFFFF)
									csleep(r->eax & 0xFFFF);
								else
									yield();
								return r;
							};
							fail = TRUE;
							break;
						}
						else if(*(uint32_t *)b == 0xFEEB4C21)
						{
							KiV86Return(r);
							return r;
						} else {
							r->useresp -= 2;
							*((uint16_t *)r->useresp) = (uint32_t)r->eflags & 0xFFFF;
							r->useresp -= 2;
							*((uint16_t *)r->useresp) = r->cs;
							r->useresp -= 2;
							*((uint16_t *)r->useresp) = (uint32_t)(r->eip + 2);
							r->eip = (void *)(uint32_t)(*(uint16_t *)b);
							r->cs = *(uint16_t *)(b+2);
							//printf("Interrupt %02X at %04X:%04X\n", *(a+1), r->cs, r->eip);
							return r;
						}
						fail = TRUE;
						break;
					case 0xCF: // IRET
						r->eip = (void *)(uint32_t)*((uint16_t *)r->useresp);
						r->useresp += 2;
						r->cs = *((uint16_t *)r->useresp);
						r->useresp += 2;
						r->eflags = (r->eflags & 0xFFFF0000) | *((uint16_t *)r->useresp);
						r->useresp += 2;
						if(r->cs == 0xFEEB)
						{
							if((uint32_t)r->eip == 0x4C21)
							{
								// for some stupid reason, dprintf fails in this function.
								// uncomment if you wish to see it.
								//printf("v86: coming home\n");
								KiV86Return(r);
								return r;
							} else {
								fail = TRUE;
								break;
							};
						};
						return r;
					case 0xE4: // IN al, __
						r->eax &= 0xFFFFFF00;
						r->eax |= inb(a[1]);
						r->eip += 2;
						return r;
					case 0xE5: // IN ax, __
						if(v86f & V86F_OPSIZE)
						{
							r->eax = ind(a[1]);
						}
						else
						{
							r->eax &= 0xFFFF0000;
							r->eax |= inw(a[1]);
						};
						r->eip += 2;
						return r;
					case 0xE6: // OUT __, al
						outb(a[1], r->eax & 0xFF);
						r->eip += 2;
						return r;
					case 0xE7: // OUT __, ax
						if(v86f & V86F_OPSIZE)
						{
							outd(a[1], r->eax);
						}
						else
						{
							outw(a[1], r->eax & 0xFFFF);
						};
						r->eip += 2;
						return r;
					case 0xEC: // IN al, dx
						r->eax &= 0xFFFFFF00;
						r->eax |= inb(r->edx & 0xFFFF);
						r->eip++;
						return r;
					case 0xED: // IN ax, dx
						if(v86f & V86F_OPSIZE)
						{
							r->eax = ind(r->edx & 0xFFFF);
						}
						else
						{
							r->eax &= 0xFFFF0000;
							r->eax |= inw(r->edx & 0xFFFF);
						};
						r->eip++;
						return r;
					case 0xEE: // OUT dx, al
						outb(r->edx & 0xFFFF, r->eax & 0xFF);
						r->eip++;
						return r;
					case 0xEF: // OUT dx, ax
						if(v86f & V86F_OPSIZE)
						{
							outd(r->edx & 0xFFFF, r->eax);
						}
						else
						{
							outw(r->edx & 0xFFFF, r->eax & 0xFFFF);
						};
						r->eip++;
						return r;
					case 0xF3: // REP
						reps = r->ecx & 0xFFFF;
						break;
					case 0xFA: // CLI
						r->eflags &= ~0x0200;
						r->eip++;
						return r;
					case 0xFB: // STI
						r->eflags |= 0x0200;
						r->eip++;
						return r;
					default:
						fail = TRUE;
				};
				a++;
			};
		};
		currtty->color = 0x1F;
		currtty->driver->cls(currtty);
		printf("\nFAULT %02X (%s), with error code %08X.\n", r->int_no, exception_messages[r->int_no], r->err_code);
		if(r->eflags & 0x20000)
		{
			printf("V86 fault occurred.\n");
			printf("CS:IP: %04X:%04X | EFLAGS: %08X\n", r->cs, r->eip, r->eflags);
		} else {
			printf("32-bit fault occurred in ring %i.\n", r->cs & 3);
			printf("CS:EIP: %04X:%08X | EFLAGS: %08X\n", r->cs, r->eip, r->eflags);
		};
		printf("EAX: %08X | EBX: %08X | ECX: %08X | EDX: %08X\n", r->eax, r->ebx, r->ecx, r->edx);
		printf("ESI: %08X | EDI: %08X | EBP: %08X | ESP: %08X\n", r->esi, r->edi, r->ebp, r->esp);
		printf("DS:  %04X     | ES:  %04X     | FS:  %04X     | GS:  %04X\n", r->ds, r->es, r->fs, r->gs);
		if(r->eflags & 0x20000)
		{
			printf("VDS: %04X     | VES: %04X     | VFS: %04X     | VGS: %04X\n", r->vds, r->ves, r->vfs, r->vgs);
		};
		tss = &ktss;
		printf("TSS-ESP: x-%04X:%08X | 2-%04X:%08X | 1-%04X:%08X | 0-%04X:%08X\n"
			, tss->ss, tss->esp, tss->ss2, tss->esp2, tss->ss1, tss->esp1, tss->ss0, tss->esp0);
		printf("USER-ESP: %04X:%08X\n", r->ss, r->useresp);
		printf("CR0: %08X | CR2: %08X | CR3: %08X | TR:  %04X\n"
			,__KiGetCR0()
			,__KiGetCR2()
			,__KiGetCR3()
			,__KiGetTR()
				);
		printf("IDUMP:");
		if(r->eflags & 0x20000)
			a = (void *)((uint32_t)r->eip + (r->cs<<4));
		else
			a = (void *)r->eip;
		for(i=0;i<16;i++)
		{
			printf(" %02X",*a);
			a++;
		}
		if(r->eflags & 0x20000) {
			a = (void *)((uint32_t)r->useresp + (r->ss<<4));
		} else if(r->cs & 3) {
			a = (void *)r->useresp;
		} else {
			a = (void *)&(r->useresp);
		};
		printf("\n");
		
		printf("STACK:\n");
		for(i=1;i<=80;i++)
		{
			printf(" %08X",*(uint32_t *)a);
			a += 4;
			if(!(i & 7))
				printf("\n");
		};
		(void)frame;
		/*dprintf("CALL STACK:\n");
		while(frame)
		{
			dprintf("eip: %p  ebp: %p\n",frame->ip, frame);
			frame = frame->prev;
		};*/
		
		if(r->eflags & 0x20000)
		{
			//panic("I can't read the last message dammit! Piss off!");
			KiV86Return(r);
			return r;
		} else {
			if((r->cs & 3) == 0x03)
			{
				__KiKillTask(currtask);
				return r;
			} else
				panic("System halted. Yes, even we make mistakes.");
		}
	};
	return r;
};

char decide(int intno)
{
	/* temporary until there is actually a reason to not return 0 */
	return 0;
	
	switch(intno)
	{
		case 0x08:
			/* double fault -- type ABORT */
			return 0;
			break;
		case 0x09:
			/* coprocessor segment overrun -- type ABORT */
			return 0;
			break;
		case 0x12:
			/* machine check -- type ABORT */
			return 0;
			break;
	};
};

void KiRestart(void)
{
	uint8_t status;
	
	asm("cli");	/* no more tasking */
	status = 0;

	do
	{
		status = inb(0x64);
		if((status & 0x01) != 0)
		{
			(void)inb(0x60);
			continue;
		};
	} while((status & 0x02) != 0);

	outb(0x64, 0xFE);
	printf("Unable to reset via keyboard. Attempting FFFF:0000 exit.\n");
	__KiRestartByJump();
	panic("Cannot restart computer?!\n");
};
