/***************************************************************
 *
 * drivers/input/i8042-kb.c
 * i8042-based keyboard driver for AWOS
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#include <portio.h>
#include <kernel.h>
#include <input.h>

void KiRestart(void);

uint32_t i8042_keymap_en[4][90] = { 
	{
		0, KEY_ESC, '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '=', '\b', '\t',
		'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '[', ']', '\n', 0, 'a', 's',
		'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', '\'', '`', 0, '\\', 'z', 'x', 'c', 'v',
		'b', 'n', 'm', ',', '.', '/', 0, 0, 0, ' ', 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, KEY_HOME, KEY_UP, KEY_PGUP, '-', KEY_LFT, 0, KEY_RT, '+', KEY_END,
		KEY_DN, KEY_PGDN, KEY_INS, KEY_DEL, 0, 0, 0, 0, 0
	},
	// SHIFT
	{
		0, 0, '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '_', '+', 0, 0,
		'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '{', '}', 0, 0, 'A', 'S',
		'D', 'F', 'G', 'H', 'J', 'K', 'L', ':', '"', '~', 0, '|', 'Z', 'X', 'C', 'V',
		'B', 'N', 'M', '<', '>', '?', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0
	},
	// NUMLOCK
	{
		0, KEY_ESC, '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '=', '\b', '\t',
		'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '[', ']', '\n', 0, 'a', 's',
		'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', '\'', '`', 0, '\\', 'z', 'x', 'c', 'v',
		'b', 'n', 'm', ',', '.', '/', 0, 0, 0, ' ', 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, '7', '8', '9', '-', '4', '5', '6', '+', '1',
		'2', '3', 0, 0, 0, 0, 0, 0, 0
	},
	// CRAPSLOCK
	{
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', 0, 0, 0, 0, 'A', 'S',
		'D', 'F', 'G', 'H', 'J', 'K', 'L', 0, 0, 0, 0, 0, 'Z', 'X', 'C', 'V',
		'B', 'N', 'M', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0
	}
};

uint8_t cstack[8]; // Key stack for processing codes like 0xE0
uint8_t cstack_idx;

uint16_t dphase;

int i8042_scan2key(int c)
{
	int k;
	
	if(c >= 90)
		return 0;
	k = i8042_keymap_en[0][c];
	if(i8042_keymap_en[1][c] != 0 && (kbd_flags & KBF_SHIFT))
	{
		if(i8042_keymap_en[3][c] == 0 || !(kbd_flags & KBF_CAPSLK))
			k = i8042_keymap_en[1][c];
		
	} else if(i8042_keymap_en[3][c] != 0 && (kbd_flags & KBF_CAPSLK))
		k = i8042_keymap_en[3][c];
	if(i8042_keymap_en[2][c] != 0 && (kbd_flags & KBF_NUMLK))
		k = i8042_keymap_en[2][c];
	
	// Some keys just have to be ignored.
	return k;
}

void i8042_kbd_irq(void *r)
{
	char key;
	(void)r;
	
	key = inb(0x60);
	
	cstack[cstack_idx++] = key;
	
	switch((unsigned char)key)
	{
		case 0xE0:
			return;
		case 0x2A:
			kbd_flags |= KBF_LSHIFT | KBF_SHIFT;
			break;
		case 0xAA:
			kbd_flags &= ~KBF_LSHIFT;
			if(!(kbd_flags & KBF_RSHIFT))
				kbd_flags &= ~KBF_SHIFT;
			break;
		case 0x36:
			kbd_flags |= KBF_RSHIFT | KBF_SHIFT;
			break;
		case 0xB6:
			kbd_flags &= ~KBF_RSHIFT;
			if(!(kbd_flags & KBF_LSHIFT))
				kbd_flags &= ~KBF_SHIFT;
			break;
		case 0x3A:
			kbd_flags ^= KBF_CAPSHD;
			if(kbd_flags & KBF_CAPSHD)
				kbd_flags ^= KBF_CAPSLK;
			break;
		case 0x45:
			kbd_flags ^= KBF_NUMHD;
			if(kbd_flags & KBF_NUMHD)
				kbd_flags ^= KBF_NUMLK;
			break;
		case 0x46:
			kbd_flags ^= KBF_SCRHD;
			if(kbd_flags & KBF_SCRHD)
				kbd_flags ^= KBF_SCRLK;
			break;
		case 0x1D:
			if(cstack_idx == 2 && cstack[0] == 0xE0)
				kbd_flags |= KBF_RCTRL;
			else
				kbd_flags |= KBF_LCTRL;
			
			kbd_flags |= KBF_CTRL;
			break;
		case 0x9D:
			if(cstack_idx == 2 && cstack[0] == 0xE0)
				kbd_flags &= ~KBF_RCTRL;
			else
				kbd_flags &= ~KBF_LCTRL;
			
			if(!(kbd_flags & (KBF_LCTRL | KBF_RCTRL)))
				kbd_flags &= ~KBF_CTRL;
			break;
		case 0x38:
			if(cstack_idx == 2 && cstack[0] == 0xE0)
				kbd_flags |= KBF_RALT;
			else
				kbd_flags |= KBF_LALT;
			
			kbd_flags |= KBF_ALT;
			break;
		case 0xB8:
			if(cstack_idx == 2 && cstack[0] == 0xE0)
				kbd_flags &= ~KBF_RALT;
			else
				kbd_flags &= ~KBF_LALT;
			
			if(!(kbd_flags & (KBF_LALT | KBF_RALT)))
				kbd_flags &= ~KBF_ALT;
			break;
		case 0x53:
			if((kbd_flags & KBF_CTRL) && (kbd_flags & KBF_ALT))
				KiRestart();
		default:
			if(key & 0x80)
			{
				
			} else {
				KbdPushKey(i8042_scan2key(key));
			};
	};
	cstack_idx = 0;
}

char i8042_command(int command, int argc)
{
	/* At the moment argc is either 1 or 0 */
	char resp = 0;

	while(inb(0x64) & 0x02)
		csleep(50);
	outb(0x64, (int8_t)command);

	if(argc)
	{
		while(~inb(0x64) & 0x01)
			csleep(50);
		resp = inb(0x60);
	};

	return resp;
}

BOOL i8042_kbd_detect(void)
{
	/* FIXME TODO */
	/* KiIrqInstall(1, i8042_kbd_irq_detect, "[DETECTING] Standard i8042 Keyboard Driver"); */
	/* dphase = 0x64AA; */
	/* if(i8042_command(0xAA, 1) == 0x55)
		return TRUE;
	
	return FALSE; */
	return TRUE;	/* A not-very-nice way to "assume" instead of detect */
}

void i8042_kbd_init(void)
{
	unsigned char ctl = i8042_command(0x20, 1);

	ctl |= 0x01; /* Set to use IRQ1 */
	ctl &= ~0x10; /* Enable */

	i8042_command(0x60, 0);
	outb(0x60, ctl);

	/* outb(0x64, 0xAE);
	csleep(10); */
	KiIrqInstall(1, i8042_kbd_irq, "Standard i8042 Keyboard Driver");
	
	return;
}
