/****************************************************************
 *
 * drivers/pci/nvidia.c
 * nVidia driver, hopefully
 * 
 * See LICENSE file included with this source for copyright
 * information.
 * 
 ****************************************************************/

#include <kernel.h>
#include <stdint.h>
#include <pci.h>
#include <nvidia.h>
#include <nvreg.h>

struct pcidev_ll *nvidia_card = NULL;

// Yay, init routine.
void InitNvidia(void)
{
	struct pcidev_ll *device;
	uint16_t addy;
	int i;
	
	for(;;)
	{
		device = PciFindByVendor(NULL, 0x10DE);
		if(device == NULL)
			break;
		if((device->dev).cls_a == 0x03 && (device->dev).cls_b == 0x00 && (device->dev).cls_c == 0x00)
			break;
	};
	
	// For testing a few things.
	//device = PciFindByClassC(NULL, 0x03, 0x00, 0x00);
	
	if(device == NULL)
		return;
	
	nvidia_card = device;
	
	printf("nvidia: nVidia video card detected\n");
	switch((device->dev).device_id)
	{
		case 0x0322: // GreaseMonkey's FX 5200 AGP
			printf("nvidia: NV34 GeForce FX 5200 AGP\n");
			break;
		case 0x0400: // awilcox's 8600 GTS
			printf("nvidia: G84 GeForce 8600 GTS\n");
			break;
		case 0x0605: // awilcox's 9800 GT
			printf("nvidia: G92 GeForce 9800 GT\n");
			break;
		default:
			printf("nvidia: card not found in database\nnvidia: please tell us the name and this: PID=%04X\n", (device->dev).device_id);
			printf("nvidia: it would also be helpul if you gave us a hex dump\n");
	};
	
	do
	{
		dprintf("nvidia: looking for HWEA signature from 000C0000 to 000C0200\n");
		// Based on awilcox's 9800 GT
		for(addy = 0x0000; addy < 0x0200; addy++)
		{
			if (*(uint32_t *)(0xC0000 | addy) == 0x41455748)
			{
				break;
			};
		};
		if(addy == 0x0200)
		{
			dprintf("nvidia: we could not find the HWEA signature\n");
		} else {
			dprintf("nvidia: HWEA signature found - attempting string read\n");
			// Now print whatever stringy thing we can find. This should be more accurate than the PCIR one.
			printf("nvidia: apparently our card is (HWEA): %s\n", (char *)(0xC0000 | addy));
			
			break; // note to froggey: no "goto"s here! :D :D :D -GM
		};
		
		dprintf("nvidia: attempting PCIR lookup from 000C0018\n");
		// OK, I'm crossing my fingers here; I cannot guarantee this will work on your card, sorry.
		// This is based on GreaseMonkey's 5200 FX AGP.
		// We are now looking into the VGA BIOS area.
		// This should point to the PCIR block.
		addy = *(uint16_t *)(0xC0018);
		// Check the signature - does it equal "PCIR"?
		if (*(uint32_t *)(0xC0000 | addy) == 0x52494350) // little-endian
		{
			dprintf("nvidia: PCIR signature found - attempting string read\n");
			// Now print whatever stringy thing we can find.
			printf("nvidia: apparently our card is (PCIR): %s\n", (char *)(0xC0000 | (addy + 0x18)));
		} else {
			dprintf("nvidia: we could not find the PCIR signature\n");
		};
	} while(0);
	
	dprintf("nvidia: bar0 (nv registers) is located at %08X\n", (device->dev).bar[0]);
	dprintf("nvidia: bar1 (LFB) is located at %08X\n", (device->dev).bar[1]);
	
	// SANITY CHECK: are the nv registers non-prefetchable?
	if((device->dev).bar[0] & 0x00000008)
	{
		i = 0;
		// XXX - i isn't defined elsewhere according to ICC. Previous declaration is to shut it up.
		printf("nvidia: Assertion! - bar0 (nv registers) is prefetchable %d vs. %d!\n", (device->dev).bar[0], (16 << 20));
		nvidia_card = NULL;
		return;
	} else {
		dprintf("nvidia: bar0 (nv registers) is non-prefetchable\n");
	};
	
	// SANITY CHECK: are the nv registers exactly 16MB?
	PciWriteConf32((int16_t)device->port, device->b, device->s, device->f, 0x10, 0xFFFFFFFF);
	i = PciReadConf32((int16_t)device->port, device->b, device->s, device->f, 0x10);
	i = ~(i & ~0xF) + 1;
	PciWriteConf32((int16_t)device->port, device->b, device->s, device->f, 0x10, (device->dev).bar[0]);
	if(i == (16 << 20))
	{
		dprintf("nvidia: bar0 (nv registers) is 16MB\n");
	} else {
		printf("nvidia: Assertion! - bar0 (nv registers) is %08X - should be 16MB (%08X)!\n", i, (16 << 20));
		nvidia_card = NULL;
		return;
	};
	
	// We must ensure that we are using the 3Dx range, not the 3Bx range.
	// Read from 3CC, then set the flag and write into 3C2.
	i = *(uint8_t *)((device->dev).bar[0] + (NV_PCIO0_OFFSET | 0x3CC)) | 0x01;
	*(uint8_t *)((device->dev).bar[0] + (NV_PCIO0_OFFSET | 0x3C2)) = (int8_t)i;
	
	// Now we shall unlock the card for read/write access.
	// nvidia register unlock
	*(uint16_t *)((device->dev).bar[0] + (NV_PCIO0_OFFSET | 0x3D4)) = 0x571F;
	// VGA CRTC register unlock
	// clear the top bit of CRTC register 0x11
	*(uint8_t *)((device->dev).bar[0] + (NV_PCIO0_OFFSET | 0x3D4)) = 0x11;
	i = ((*(uint8_t *)((device->dev).bar[0] + (NV_PCIO0_OFFSET | 0x3D5)) << 8) & 0x7F) | 0x11;
	*(uint16_t *)((device->dev).bar[0] + (NV_PCIO0_OFFSET | 0x3D4)) = (int16_t)i;
	
	printf("nvidia: card should be unlocked; read/write access allowed\n");
	
	// OK, let's try reading the revision...
	*(uint8_t *)((device->dev).bar[0] + (NV_PCIO0_OFFSET | 0x3D4)) = 0x27;
	i = ((*(uint8_t *)((device->dev).bar[0] + (NV_PCIO0_OFFSET | 0x3D5)) << 8) & 0x7F);
	
	printf("nvidia: apparently the card revision is %02X\n", i);
	
	printf("nvidia: RAMHT = %08X, RAMFC = %08X\n"
		,*(uint8_t *)((device->dev).bar[0] + 0x002210)
		,*(uint8_t *)((device->dev).bar[0] + 0x002214)
			);
	
	return;
}
